////////////////////////////////////////////////////////////////////////////
// NoteCase notes manager project <http://notecase.sf.net>
//
// This code is licensed under BSD license.See "license.txt" for more details.
//
// File: Implements Blowfish encrypted NoteCase HTML format I/O
////////////////////////////////////////////////////////////////////////////

#include "FormatIOEncHtml.h"
#include "IOProcess.h"
#include "IOLayerZlib.h"
#include "IOLayerEnc.h"
#include "IOLayerRedirect.h"
#include "File64.h"
#include "IOLayerFile64.h"
#include "DocumentIterator.h"
#include "SHA1.h"
#include "debug.h"
#include <stdio.h>	//sprintf
#include <glib.h>	// bate order macros

void WriteLevel(DocumentIterator &it, IOProcess &file, int nParentID);

void ParseRedirector(const char *, int);
static HTMLParser *g_pParser = NULL;

FormatIO_EncHTML::FormatIO_EncHTML()
{
	m_nOrigSize = 0;
	m_bNewFormat = true; //default
}

FormatIO_EncHTML::~FormatIO_EncHTML()
{
}

void FormatIO_EncHTML::SetPassword(const char *szPass)
{
	m_strPass = szPass;
}

int FormatIO_EncHTML::Load(const char *szFile, NoteDocument &doc)
{
	if(m_strPass.size()==0)
		return DOC_LOAD_WRONG_PASSWORD;	//password not set

	IOProcess objFile(MODE_READ);

	// attach file read layer
	IOLayerFile64 *pFileLayer = new IOLayerFile64;
	objFile.PushLayer(pFileLayer);	

	// attach content decryption layer
	IOLayerEnc *pEncLayer = new IOLayerEnc;
	pEncLayer->SetPassword(m_strPass.c_str());
	objFile.PushLayer(pEncLayer);	// decryption step

	char szBuffer[1032];
	int nRead;

	m_objDoc = &doc;  //store pointer

	//clear old document contents
	doc.Clear();

	if(!pFileLayer->m_file.Open(szFile, F64_READ|F64_SHARE_READ|F64_OPEN_EXISTING))
		return DOC_LOAD_NOT_FOUND;

	//
	// read encrypted format header
	//
	
	// 1. format signature text "NOTECASE" (8 bytes)
	char szFormatCode[9];
	pFileLayer->m_file.Read(szFormatCode, 8);	// "explicit qualification"
	szFormatCode[8] = '\0';
	if(0 != strcmp(szFormatCode, "NOTECASE"))
		return DOC_LOAD_FORMAT_ERROR;	//
	
	//---- FORMAT VERSION DETECTION ----
	// 2. test for new version format - read version string (4 bytes)
	char szFormatVer[5];
	pFileLayer->m_file.Read(szFormatVer, 4);	// "explicit qualification"
	szFormatVer[4] = '\0';
	if(0 != strcmp(szFormatVer, "enc2"))
	{
		m_bNewFormat = false;
		pFileLayer->m_file.Seek(8);	// roll back
	}

	// if new format attach content decompression layer
	IOLayerZlib *pZlibLayer = NULL;
	if(m_bNewFormat)
	{
		pZlibLayer = new IOLayerZlib;
		objFile.PushLayer(pZlibLayer);
	}

	// attach redirector object
	IOLayerRedirect *pRedirectLayer = new IOLayerRedirect;
	pRedirectLayer->SetCallback(ParseRedirector);
	g_pParser = this;
	objFile.PushLayer(pRedirectLayer);

	// 3. SHA-1 password hash value (20 bytes)
	char szHashRead[20] = "";
	pFileLayer->m_file.Read(szHashRead, 20);
	CSHA1 hash;
	hash.Update((BYTE *)m_strPass.c_str(), m_strPass.size());
	hash.Final();
	if(0 != memcmp(szHashRead, hash.m_digest, 20))
		return DOC_LOAD_WRONG_PASSWORD;	//invalid password
	
	// 4. original (unencrpyted) file size (6 bytes)
	pFileLayer->m_file.Read((char *)&m_nOrigSize, sizeof(m_nOrigSize));

	// 5. compressed stream size (8 bytes)
	if(m_bNewFormat){
		pFileLayer->m_file.Read((char *)&(pZlibLayer->m_nTotalInBytes), sizeof(INT64));//TOFIX update at the end

		//convert from network byte order (big-endian) to processor default
		if(G_BYTE_ORDER == G_LITTLE_ENDIAN)
			pZlibLayer->m_nTotalInBytes = GINT64_FROM_BE(pZlibLayer->m_nTotalInBytes);

		TRACE("Zlib stream size as read from file is: %d\n", (int)(pZlibLayer->m_nTotalInBytes)); //TOFIX printf flag for int64!
	}

	//
	// main reading loop
	//  (some layers cache data from one step and can return 0 bytes in some cycle)
	//TOFIX error handling
	bool bContinue = true;
	while(bContinue)
		nRead = objFile.Read(szBuffer, 1024, bContinue);
	objFile.Finalize();

	if(m_bNewFormat)
	{
		TRACE("Zlib decompress total bytes input: %d\n", (int)(pZlibLayer->m_nProcessedInBytes)); //TOFIX printf flag for int64!
		TRACE("Zlib decompress total bytes output: %d\n", (int)(pZlibLayer->m_nProcessedOutBytes)); //TOFIX printf flag for int64!
	}

	//TOFIX use m_nOrigSize
	return DOC_LOAD_OK;
}

bool FormatIO_EncHTML::Save(const char *szFile, NoteDocument &doc)
{
	if(m_strPass.size()==0)
		return false;	//password not set

	IOProcess objFile(MODE_WRITE);

	// attach content compression layer
	IOLayerZlib *pZlibLayer = NULL;
	if(m_bNewFormat){
		pZlibLayer = new IOLayerZlib;
		objFile.PushLayer(pZlibLayer);	// compress before encrypting
	}
	
	// attach content encryption layer
	IOLayerEnc *pEncLayer = new IOLayerEnc;
	pEncLayer->SetPassword(m_strPass.c_str());
	objFile.PushLayer(pEncLayer);	// encryption step

	// attach file write layer
	IOLayerFile64 *pFileLayer = new IOLayerFile64;
	objFile.PushLayer(pFileLayer);	
		
	if(!pFileLayer->m_file.Open(szFile, F64_WRITE|F64_SHARE_READ|F64_OPEN_NEW))
		return false;

	//
	// write encrypted format header
	//
	
	// 1. format signature text "NOTECASE" (8 bytes)
	pFileLayer->m_file.Write("NOTECASE", 8);

	// 2. write format version string (4 bytes) - new format
	if(m_bNewFormat)
		pFileLayer->m_file.Write("enc2", 4);

	// 3. SHA-1 password hash value (20 bytes)
	CSHA1 hash;
	hash.Update((BYTE *)m_strPass.c_str(), m_strPass.size());
	hash.Final();
	pFileLayer->m_file.Write((char *)hash.m_digest, 20);
		
	// 4. original (unencrpyted) file size (8 bytes)
	pFileLayer->m_file.Write((char *)&m_nOrigSize, sizeof(m_nOrigSize));//TOFIX update at the end
	
	// 5. compressed stream size (8 bytes)
	if(m_bNewFormat)
	{
		INT64 nSize = 0;
		pFileLayer->m_file.Write((char *)&(nSize), sizeof(INT64));
	}

	//
	// Start writing the main part of the stream
	//

	//write header
	objFile.WriteString("<!DOCTYPE NoteCenter-File>\r\n"); //TOFIX

	//calc and write selected node "index"
	char szLastNote[100];
	sprintf(szLastNote, "<!--LastNote:%d-->\r\n", doc.m_nCurrentNode);
	objFile.WriteString(szLastNote);
	
	objFile.WriteString("<HTML>\r\n<HEAD>\r\n<TITLE>");
	//TOFIX write file title as document title
	objFile.WriteString("</TITLE>\r\n</HEAD>\r\n<BODY BGCOLOR=\"#FFFFFF\">\r\n");

	//recursively iterate through document node tree
	DocumentIterator it(doc);
	WriteLevel(it, objFile, -1);

	//write footer
	objFile.WriteString("</BODY>\r\n</HTML>\r\n");

	objFile.Finalize();
	
	//TOFIX update m_nOrigSize

	// update item #5. compressed stream size (8 bytes)
	if(m_bNewFormat)
	{
		TRACE("Zlib stream bytes input: %d / output: %d\n", (int)(pZlibLayer->m_nProcessedInBytes), (int)(pZlibLayer->m_nProcessedOutBytes)); //TOFIX printf flag for int64!
		pFileLayer->m_file.Seek(40);

		//convert from processor default to network byte order (big-endian)
		INT64 nSize = pZlibLayer->m_nProcessedOutBytes;
		if(G_BYTE_ORDER == G_LITTLE_ENDIAN)
			nSize = GINT64_TO_BE(nSize);

		pFileLayer->m_file.Write((char *)&(nSize), sizeof(INT64));
	}

	return true;
}

void ParseRedirector(const char *buf, int len)
{
	if(g_pParser)
		g_pParser->Parse(buf, len);
}
